@@ -7,14 +7,16 @@ if sys.version_info < (2, 7):
 else:
     import unittest  # pragma: nocover
 
+import webob
+from webob.exc import HTTPNotFound
 from webtest import TestApp
 import six
 from six import b as b_
 from six.moves import cStringIO as StringIO
 
 from pecan import (
-    Pecan, expose, request, response, redirect, abort, make_app,
-    override_template, render
+    Pecan, Request, Response, expose, request, response, redirect,
+    abort, make_app, override_template, render
 )
 from pecan.templating import (
     _builtin_renderers as builtin_renderers, error_formatters
@@ -46,7 +48,8 @@ class TestEmptyContent(PecanTestCase):
 
     def test_empty_index(self):
         r = self.app_.get('/')
-        self.assertEqual(r.status_int, 200)
+        self.assertEqual(r.status_int, 204)
+        self.assertNotIn('Content-Type', r.headers)
         self.assertEqual(r.headers['Content-Length'], '0')
         self.assertEqual(len(r.body), 0)
 
@@ -283,7 +286,7 @@ class TestControllerArguments(PecanTestCase):
                 )
 
             @expose()
-            def _route(self, args):
+            def _route(self, args, request):
                 if hasattr(self, args[0]):
                     return getattr(self, args[0]), args[1:]
                 else:
@@ -758,6 +761,42 @@ class TestControllerArguments(PecanTestCase):
         assert r.body == b_('eater: 10, dummy, day=12, month=1')
 
 
+class TestDefaultErrorRendering(PecanTestCase):
+
+    def test_plain_error(self):
+        class RootController(object):
+            pass
+
+        app = TestApp(Pecan(RootController()))
+        r = app.get('/', status=404)
+        assert r.status_int == 404
+        assert r.content_type == 'text/plain'
+        assert r.body == b_(HTTPNotFound().plain_body({}))
+
+    def test_html_error(self):
+        class RootController(object):
+            pass
+
+        app = TestApp(Pecan(RootController()))
+        r = app.get('/', headers={'Accept': 'text/html'}, status=404)
+        assert r.status_int == 404
+        assert r.content_type == 'text/html'
+        assert r.body == b_(HTTPNotFound().html_body({}))
+
+    def test_json_error(self):
+        class RootController(object):
+            pass
+
+        app = TestApp(Pecan(RootController()))
+        r = app.get('/', headers={'Accept': 'application/json'}, status=404)
+        assert r.status_int == 404
+        json_resp = json.loads(r.body.decode())
+        assert json_resp['code'] == 404
+        assert json_resp['description'] is None
+        assert json_resp['title'] == 'Not Found'
+        assert r.content_type == 'application/json'
+
+
 class TestAbort(PecanTestCase):
 
     def test_abort(self):
@@ -936,6 +975,57 @@ class TestStreamedResponse(PecanTestCase):
         assert r.body == b_('plain text')
 
 
+class TestManualResponse(PecanTestCase):
+
+    def test_manual_response(self):
+
+        class RootController(object):
+            @expose()
+            def index(self):
+                resp = webob.Response(response.environ)
+                resp.body = b_('Hello, World!')
+                return resp
+
+        app = TestApp(Pecan(RootController()))
+        r = app.get('/')
+        assert r.body == b_('Hello, World!')
+
+
+class TestCustomResponseandRequest(PecanTestCase):
+
+    def test_custom_objects(self):
+
+        class CustomRequest(Request):
+
+            @property
+            def headers(self):
+                headers = super(CustomRequest, self).headers
+                headers['X-Custom-Request'] = 'ABC'
+                return headers
+
+        class CustomResponse(Response):
+
+            @property
+            def headers(self):
+                headers = super(CustomResponse, self).headers
+                headers['X-Custom-Response'] = 'XYZ'
+                return headers
+
+        class RootController(object):
+            @expose()
+            def index(self):
+                return request.headers.get('X-Custom-Request')
+
+        app = TestApp(Pecan(
+            RootController(),
+            request_cls=CustomRequest,
+            response_cls=CustomResponse
+        ))
+        r = app.get('/')
+        assert r.body == b_('ABC')
+        assert r.headers.get('X-Custom-Response') == 'XYZ'
+
+
 class TestThreadLocalState(PecanTestCase):
 
     def test_thread_local_dir(self):
@@ -1005,7 +1095,7 @@ class TestFileTypeExtensions(PecanTestCase):
 
     def test_hidden_file(self):
         r = self.app_.get('/.vimrc')
-        assert r.status_int == 200
+        assert r.status_int == 204
         assert r.body == b_('')
 
     def test_multi_dot_extension(self):
@@ -1501,3 +1591,40 @@ class TestEngines(PecanTestCase):
         r = app.get('/')
         assert r.status_int == 200
         assert b_("<h1>Hello, Jonathan!</h1>") in r.body
+
+    def test_default_json_renderer(self):
+
+        class RootController(object):
+            @expose()
+            def index(self, name='Bill'):
+                return dict(name=name)
+
+        app = TestApp(Pecan(RootController(), default_renderer='json'))
+        r = app.get('/')
+        assert r.status_int == 200
+        result = dict(json.loads(r.body.decode()))
+        assert result == {'name': 'Bill'}
+
+
+class TestDeprecatedRouteMethod(PecanTestCase):
+
+    @property
+    def app_(self):
+        class RootController(object):
+
+            @expose()
+            def index(self, *args):
+                return ', '.join(args)
+
+            @expose()
+            def _route(self, args):
+                return self.index, args
+
+        return TestApp(Pecan(RootController()))
+
+    def test_required_argument(self):
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            r = self.app_.get('/foo/bar/')
+            assert r.status_int == 200
+            assert b_('foo, bar') in r.body
